<?php defined('SYSPATH') OR die('No direct access allowed.');

require_once('op5/log.php');

class Synergy_Model extends Model
{
	protected static $_synergy_path = "/opt/synergy/";

	private static function log($level, $msg) {
		$log = Op5Log::instance('synergy');
		$log->log($level, $msg);
	}

	/**
	 *	Fetch a list of all saved trees
	 *	@return array
	 */
	static function getRuleList() {
		$result = array();
		$rules_dir = self::$_synergy_path . "etc/bps";
		if (!is_dir($rules_dir)) {
			throw new SynergyException("$rules_dir is not a directory.");
		}
		if (false === ($dh = opendir($rules_dir))) {
			throw new SynergyException("Can not open directory $rules_dir.");
		}
		while(false !== ($file = readdir($dh))) {
			if(($file[0] === ".") || pathinfo($file, PATHINFO_EXTENSION) !== "lua") continue;
			$file = urldecode($file);
			$result[] = substr($file, 0, -4);
		}
		natcasesort($result);
		// Javascript prints the DOM with reverse sort for some reason. Will find out why later
		return array_reverse($result, true);
	}

	// Checks if user is permitted to see the service.
	private static function __is_permitted_by_services($data, $auth) {
		if ($auth->authorized_for('service_view_all')) {
			return true;
		}

		// Service can be seen by all users.
		if ($data["result"]["service_contacts"] === "*") {
			return true;
		}

		// Users is contact for service
		if (in_array($auth->get_user()->username, $data["result"]["service_contacts"])) {
			return true;
		}

		return false;
	}

	// Checks if user is permitted to see the host.
	private static function __is_permitted_by_hosts($data, $auth) {
		if ($auth->authorized_for('host_view_all')) {
			return true;
		}

		// Host can be seen by all users.
		if ($data["result"]["host_contacts"] === "*") {
			return true;
		}

		// User is contact for host
		if (in_array($auth->get_user()->username, $data["result"]["host_contacts"])) {
			return true;
		}
		return false;
	}

	/**
	 *	Pretty printing JSON
	 *	@param $block array
	 *	@param $indent int
	 *	@return string
	 */
	public static function _block_format(array $block, $ident = 1) {
		$result = "{\n";
		$i = str_repeat("\t", $ident);
		$c = 0;
		$size = sizeof($block);
		foreach($block as $key => $value) {
			$sep = ++$c == $size ? "" : ",";
			if(is_array($value)) {
				$add = "";
				if(!is_numeric($key)) {
					$add = $key . " = \n" . str_repeat("\t", $ident);
				}
				$result .= sprintf("$i$add%s$sep\n", self::_block_format($value, $ident + 1));
				continue;
			}
			if($key === "intent") {
				$result .= sprintf("$i%s = %s$sep\n", $key, $value);
				continue;
			}
			$result .= sprintf("$i%s = \"%s\"$sep\n", $key, $value);
		}
		return $result . str_repeat("\t", $ident - 1). "}";
	}

	/**
	 *	Saves created rule.
	 *	@param $rule array
	 *	@param $rule_name string
	 *	@param $sid string
	 *	@return bool
	 */
	public static function saveRule(array $rule, $rule_name, $sid) {
		$file_name = urlencode($rule_name);
		$dir_path = self::$_synergy_path . "var/queue/$sid";
		if (!is_dir($dir_path)) {
			mkdir($dir_path, 0777, true);
		}
		// Pretty prints JSON and puts content into queue
		return @ file_put_contents("$dir_path/$file_name.lua", "script = " . self::_block_format($rule));
	}

	/**
	 *	Move saved rules to spool directory
	 *	@param @sid string
	 *	@param string
	 *	@return bool(false) or string
	 */
	public static function commitActions($sid)
	{
		$auth = Auth::instance();
		$dir_path = self::$_synergy_path . "var/queue/$sid";
		if (!is_dir($dir_path)) {
			self::log('error', "Directory " . $dir_path . " doesn't exist");
			return false;
		}
		if (!$auth->authorized_for('export')) {
			self::log('error', $auth->get_user()->username . ' is not allowed to save configuration.');
			return false;
		}
		system("mv -f $dir_path ".self::$_synergy_path."var/spool/");
		// Trigger the spool to verify and enforce the changes, exec here, we don't need the output
		$return_code = exec(self::$_synergy_path . "bin/synergy --process_spool");

		// Make synergy processor process the states, to make sure the config is imported.
		self::daemonRequest('forceloop');

		return $return_code;
	}

	/**
	 *	Deletes saved rule
	 *	@param $rule_name string
	 *	@return bool
	 */
	public static function deleteRule($rule_name) {
		$file_name = urlencode($rule_name);
		$rule_path = self::$_synergy_path . "etc/bps/$file_name.lua";
		if (!file_exists($rule_path)) {
			self::log('error', 'Rule ' . $rule . ' wanted for deletion does not exist. Huh, an error with desired result...');
			return false;
		}
		@ unlink($rule_path);
		return true;
	}

	/**
	 *	Deletes list of saved rules
	 *	@param $list array
	 *	@param $sid string
	 *	@return bool(false) or string
	 */
	public static function deleteRulesList($list, $sid) {
		foreach ($list as $key => $rule) {
			$file_name = urlencode($rule);
			if (!file_exists(self::$_synergy_path . "etc/bps/".$file_name.".lua")) {
				self::log('error', 'Rule ' . $rule . ' wanted for deletion does not exist. Huh, an error with desired result...');
				return false;
			}
			$list[$key] = $file_name;
		}
		$dir_path = self::$_synergy_path . "var/queue/$sid";
		if (!is_dir($dir_path)) {
			mkdir($dir_path, 0777, true);
		}
		return @ file_put_contents("$dir_path/delete.list", json_encode(array_values($list)));
	}

	/**
	 * Send a command to daemon
	 *
	 * @param $cmd string
	 * @return array (php version of JSON result from synergy binary)
	 **/
	static function daemonRequest($cmd) {
		/* FIXME: don't hardcode the synergy socket path */
		$socket = @fsockopen('unix:///opt/monitor/var/rw/synergy', NULL, $errno, $errstr);
		if(!$socket) {
			self::log('error', 'Error connecting to daemon socket ('.$errno.'): ' . $errstr);
			throw new SynergyException("Could not connect to Synergy daemon. Is the synergy daemon running? Try run: \"service synergy restart\" from the op5 monitor system command line");
		}

		fwrite($socket, $cmd . "\n");
		$json = "";
		do {
			$response = "";
			$response = fread($socket, 2048);
			$json .= $response;
		} while ($response !== "");
		fclose($socket);
		return $json;
	}

	/**
	 * Json decode that catches errors with the decoding
	 *
	 * @param $json string
	 * @throws SynergyException
	 * @return array
	 **/
	function decode_json_to_array($json)
	{
		$data = json_decode($json, true);
		if ($data === null) {
			$jerr = json_last_error();
			self::log('error', 'Invalid return from json_decode() (' . var_export($data, true) . "). error=" . $jerr . ", input=" . $json);
			throw new SynergyException("json_decode() failed");
		}
		return $data;
	}

	/**
	 * Fetch structure and states for all trees
	 *
	 * @return array
	 */
	static function getStates() {
		$json = self::daemonRequest("full");
		$data = self::decode_json_to_array($json);
		// FIXME! Authorize materialized objects.
		$auth = Auth::instance();
		$result = array();
		foreach ($data['items'] as $key => $rule) {
			unset($data['items'][$key]);
			$data['items'][urldecode($key)] = $rule;
			if (!self::__is_permitted_by_hosts($rule, $auth) && !self::__is_permitted_by_services($rule, $auth)) {
				unset($data['items'][urldecode($key)]);
			}
		}
		uksort($data['items'], 'strnatcasecmp');
		$result['items'] = array_reverse($data['items'], true);
		return $result;
	}

	/**
	 * Fetch top level results.
	 *
	 * @todo: Uses some sort of funky authentication which must be overseen
	 * @throws SynergyException
	 * @return array
	 */
	static function getTopsList() {
		$json = self::daemonRequest("top");
		$data = self::decode_json_to_array($json);
		$auth = Auth::instance();
		$result = array();
		foreach($data['items'] as $rule => $content) {
			if(self::__is_permitted_by_services($content, $auth)) {
				$result[urldecode($rule)] = $content;
			}
		}
		uksort($result, 'strnatcasecmp');
		return array_reverse($result, true);
	}

	/**
	 *	Checks if given rule exists
	 *	@param $rule string
	 *	@return bool
	 */
	public static function ruleExists($rule) {
		$rule = urlencode($rule);
		return file_exists(self::$_synergy_path."etc/bps/$rule.lua");
	}

	/**
	 *	Fetches the different models that can be used to create rules. i.e best_state etc.
	 *	@return string
	 */
	public static function getActionsJson() {
		$actions_path = self::$_synergy_path . "etc/rules/repo";
		if(!is_dir($actions_path))
			throw new Exception("The rule template directory doesn't exist." . $actions_path);
		if(!($dir = opendir($actions_path)))
			throw new Exception("Cannot open rule template directory. " . $actions_path);
		$result = array();
		while(($file = readdir($dir)) !== false) {
			if($file[0] !== "." && pathinfo($file, PATHINFO_EXTENSION) === "lua") {
				$result[] = substr($file, 0, -4);
			}
		}
		$custom_actions_path = self::$_synergy_path . "etc/rules/custom/repo";
		if(!is_dir($custom_actions_path)) return false;
		if(!($dir = opendir($custom_actions_path))) return false;
		while(($file = readdir($dir)) !== false) {
			if($file[0] !== "." && array_pop(explode(".", $file)) === "lua") {
				$result[] = substr($file, 0, -4);
			}
		}
		$actions = addslashes(json_encode(array_unique($result)));
		$json = array();
		$status = 0;
		exec("echo $actions | " . self::$_synergy_path . "bin/actions --all 2>&1", $json, $status);
		if($status !== 0) {
			self::log("error", "[$status] Syntax error in a rule file: ".var_export($json, true));
			return false;
		}
		return implode("",$json);
	}

	/**
	 *	Fetch all nagios object names of type (hosts, services, hostgroups, serviegroups) or ALL objects if no type is specified
	 *	@param $object_type string
	 *	@param $encoded bool
	 *	@return array
	 */
	public static function getNagiosObjectsJson($object_type = false, $encoded = true) {
		$ls = Livestatus::instance();
		$result = array();
		switch ($object_type) {
			case 'hosts':
				$hosts = $ls->getHosts(array('columns'=>array('name')));
				foreach ($hosts as $host) {
					$result[$host['name']] = $host['name'];
				}
				break;
			case 'services':
				$services = $ls->getServices(array('columns'=>array("host_name", "description")));
				foreach ($services as $host_desc) {
					$host = $host_desc['host_name'];
					$desc = $host_desc['description'];
					if (!isset($result[$host])) {
						$result[$host] = array();
					}
					array_push($result[$host], $desc);
				}
				break;
			case 'hostgroups':
				$hostgroups = $ls->getHostgroups(array('columns'=>array("name")));
				foreach ($hostgroups as $hostgroup) {
					$result[$hostgroup['name']] = $hostgroup['name'];
				}
				break;
			case 'servicegroups':
				$servicegroups = $ls->getServicegroups(array('columns'=>array("name")));
				foreach ($servicegroups as $servicegroup) {
					$result[$servicegroup['name']] = $servicegroup['name'];
				}
				break;
			default:
				// Fetch_all. This shouldn't be used for synergy. Fetch on demand instead
				$result['hosts'] = self::getNagiosObjectsJson('hosts', false);
				$result['services'] = self::getNagiosObjectsJson('services', false);
				$result['hostgroups'] = self::getNagiosObjectsJson('hostgroups', false);
				$result['servicegroups'] = self::getNagiosObjectsJson('servicegroups', false);
				return json_encode($result);
				break;
		}
		if (!$encoded) {
			return $result;
		} else {
			return json_encode($result);
		}
	}
}
