-- Objects can only be of these types
local allowed_tables = {
	hosts = true,
	services = true,
	hostgroups = true,
	servicegroups = true,
	contacts = true,
	contactgroups = true
}

states = {}

local status, res = pcall(require, "config")
if not status then
	SYSError("Failed to load config: "..(res or "").." Synergy can't run without config. Exiting.")
	-- Can't continue without config. Die
	os.exit(2)
end

-- Cached contactgroups
contactgroups_cache = {}

local Request_mt = {
	compile = function(self)
		local res = string.format("GET %s\n", self.type or "hosts")
		if self.where_stmt then
			for i, v in pairs(self.where_stmt) do
				res = res .. string.format("Filter: %s\n", v)
			end
		end
		return res
	end,
	where = function(self, str, ...)
		str = string.gsub(str, "?", function (w) return "%s" end)
		self.where_stmt = self.where_stmt or {}
		table.insert(self.where_stmt, string.format(str, ...))
		return self
	end
}

Request = {
	new = function(type)
		type = type or "hosts"
		if not allowed_tables[type] then
			SYSLog("Wrong request type", 0)
			return nil, "Wrong request type"
		end
		local new_inst = { type = type }
		setmetatable(new_inst,{ __index = Request_mt })
		return new_inst
	end
}

mklive = {
	socket = nil,
	init_socket = function(self)
		local status, res, msg
		local tries = 0
		while true do
			local ss
			ss, self.socket = pcall(socket.unix)
			if ss then
				s, socket_path = pcall(get_socket_path)
				status, res, msg = pcall(self.socket.connect, self.socket, socket_path)
				if not status then
					SYSError("Livestatus socket: Failed to connect to socket")
					tries = tries + 1
					if tries >= config.max_retries then
						SYSDebug("No connection to Livestatus socket after "..config.max_retries.." retries. Skipping object.")
						SYSError("Synergy encountered errors with the Livestatus socket. A \"mon restart\" might fix this.")
						break
					end
					sleep(config.wait_on_fail)
				else
					break
				end
			else
				sleep(config.wait_on_fail)
			end
		end
	end,
	request = function(self, req)
		req = req:compile()
		req = req .. "OutputFormat: json\nKeepAlive: on\nColumnHeaders: on\nResponseHeader: fixed16\n\n"
		local status, msg, res
		local st = socket.gettime()
		local tries = 0
		while true do
			status, res, msg = pcall(function(self)
				self.socket:send(req)
				local response = self.socket:receive(16)
				if not response then
					error()
				end
				local code_status = tonumber(response:sub(1,3))
				local content_len = tonumber(response:sub(5,15))
				local response_content = self.socket:receive(content_len)
				if code_status ~= 200 then
					error {is_livestatus = true, msg = "["..code_status.."] "..response_content}
				end
				return response_content
			end, self)
			if not status then
				tries = tries + 1
				if tries >= config.max_retries then
					SYSDebug("No response from Livestatus socket after "..config.max_retries.." retries. Skipping object.")
					SYSError("Synergy encountered errors with the Livestatus socket. A \"mon restart\" might fix this.")
					-- Set global skip to enforce skipping of entire loop. Unset at end of loop.
					skip_rest_of_loop = true
					break
				end
				res = res or ""
				if res.is_livestatus then
					SYSError("Livestatus query: "..res.msg)
					error()
				else
					self.socket:close()
					SYSDebug("sleep before socket recall.")
					SYSError("Livestatus socket: "..res)
					sleep(config.wait_on_fail)
					self:init_socket()
				end
			else
				break
			end
		end
		if type(res) ~= "string" then
			SYSDebug("Wrong request response " .. json.encode(res))
		end
		return res
	end
}

-- Service metatable function
local Service_mt = {}
-- Host metatable function
local Host_mt = {
	service = function(self, name)
		if not skip_rest_of_loop then
			local new_service = response_to_obj(mklive:request(Request.new("services"):where("host_name = ?", self.name):where("description = ?", name)), Service_mt, self.name, name)
			return new_service
		end
	end
}
function host(name)
	if not skip_rest_of_loop then
		local new_host = response_to_obj(mklive:request(Request.new("hosts"):where("name = ?", name)), Host_mt, name)
		return new_host
	end
end
-- Hostgroup metatable function
local Hostgroup_mt = {
	service = function(self, name)
		if not skip_rest_of_loop then
			local new_service = response_to_obj(mklive:request(Request.new("services"):where("hostgroup_name = ?", self.name):where("description = ?", name)), Service_mt, self.name, name)
			return new_service
		end
	end
}
function hostgroup(name)
	if not skip_rest_of_loop then
		local new_hostgroup = response_to_obj(mklive:request(Request.new("hostgroups"):where("hostgroup_name = ?", name)), Hostgroup_mt, name)
		return new_hostgroup
	end
end
-- Servicegroup metatable function
local Servicegroup_mt = {}
function servicegroup(name)
	if not skip_rest_of_loop then
		local new_servicegroup = response_to_obj(mklive:request(Request.new("servicegroups"):where("servicegroup_name = ?", name), nil, name), Servicegroup_mt)
		return new_servicegroup
	end
end

rule = { type = "rule" }
nagios = { type = "nagios" }
local rule_cache = {}
local mt = {
	__index = function(self, name, ...)
		if type(rule_cache[name]) == "nil" then
			local status, action, msg = pcall(require, "repo/"..name)
			if not status then
				SYSError("Require action "..name.." load failure. "..action)
				error()
			else
				rule_cache[name] =  action[name]
			end
		end
		return function(...)
			return { args = pack_arguments(...), func = rule_cache[name].rule, name = name, type = self.type }
		end
	end
}
setmetatable(rule, mt)
setmetatable(nagios, mt)
