#!/usr/bin/lua
-- Contains all functions that processor uses

----------------------------------------------
-- Check if machine is a poller
-- @return bool
----------------------------------------------
function is_poller()
	return (os.execute("(/usr/bin/mon node show --type=master|/bin/grep -q 'TYPE=master') 2> /dev/null", 'r') == 0)
end

----------------------------------------------
-- Get a random number
-- @param vararg
-- @return int
----------------------------------------------
function real_random(...)
	return math.random(...)
end

----------------------------------------------
-- Sleep for supplied seconds
-- @param sec int
-- @return void
----------------------------------------------
function sleep(sec)
	socket.select(nil, nil, sec)
end

----------------------------------------------
-- Documented function
-- @param res string
-- @param inherits metatable function
-- @return table
----------------------------------------------
function response_to_obj(res, inherits, ...)
	local s = socket.gettime()
	if type(res) ~= "string" then
		SYSError("Failed unserializing livestatus argument")
		return nil
	end
	local status, res, msg = pcall(json.decode, res)
	if not status then
		SYSError("Failed decoding livestatus json")
		return nil
	end
	local inst = {}
	if #res < 2 then
		return nil, "Requested object not found"
	end
	if inherits then
		setmetatable(inst, { __index = inherits })
	end
	for i, v in pairs(res[1]) do
		inst[v] = res[2][i]
	end
	return inst
end

----------------------------------------------
-- Log to syslog fi config.debug is true
-- @param msg string
-- @return void
----------------------------------------------
function SYSDebug(msg, ...)
	if not config.debug then
		return
	end
	log.syslog("LOG_NOTICE", msg, ...)
	print(os.time(), "DEBUG ", msg, ...)
end

----------------------------------------------
-- Log errors to syslog and stderr
-- @param msg string
-- @return void
----------------------------------------------
function SYSError(msg, ...)
	log.syslog("LOG_ERR", msg, ...)
	io.stderr:write(msg)
end

----------------------------------------------
-- Clear table of cached states
-- @return void
----------------------------------------------
function clear_state_table()
	SYSDebug("Clearing state table.")
	states = {}
end

----------------------------------------------
-- Gets configurable livestatus socket path
-- @return string
----------------------------------------------
function get_socket_path()
	local socket_path = nil
	local file = io.open(config.livestatus_config_path, "r")
	if file then
		-- Get configured socket path
		for line in file:lines() do
			if string.sub(line, 0, 5) == "path:" then
				socket_path = string.match(line, "[^%s]+/.*")
			end
		end
		file:close()
	end
	-- Default
	if not socket_path then
		socket_path = config.default_ls_path
	end
	return socket_path
end

----------------------------------------------
-- Process all changes to trees
-- @return void
----------------------------------------------
function process_rules_spool_all()
	for uuid in lfs.dir(config.spool_dir) do
		if (uuid ~= "." and uuid ~= "..") then
			if (lfs.attributes(config.spool_dir..uuid).mode == "directory") then
				process_rules_spool(uuid)
				os.execute("rm -Rf "..config.spool_dir..uuid)
			end
		end
	end
	SYSDebug("Finished checking for spool files")
end

----------------------------------------------
-- Process changes to trees in specific folder
-- @param uuid string
-- @return void
----------------------------------------------
function process_rules_spool(uuid)
	local c = 0
	for file in lfs.dir(config.spool_dir..uuid) do
		if (file ~= "." and file ~= ".." and file:sub(-4) == ".lua") then
			local s, m = os.execute("mv -f \""..config.spool_dir..uuid.."/"..file.."\" "..config.tree_dir)
			c = 1
			if not s then
				SYSError("Cannot move new "..file:sub(0, -5).." rule in working dir: "..m)
			else
				SYSDebug("New rule "..file:sub(0, -5).." successfully loaded.")
			end
		end
	end
	if c == 0 then
		SYSDebug("Spool folder empty for new top level elements.")
	end
	local del_list = io.open(config.spool_dir..uuid.."/delete.list")
	if not del_list then
		SYSDebug("No delete.list file in spool dir.")
		return
	end
	local s, r = pcall(del_list.read, del_list, "*all")
	if not s then
		SYSError("Spool delete list content reading error. List ignored.")
		return
	end
	s, r = pcall(json.decode, r)
	if not s then
		SYSError("Spool delete list json decoding error. List ignored.")
		return
	end
	for i, v in ipairs(r) do
		if type(v) == "string" then
			local s, m = os.execute("rm -f \""..config.tree_dir..v..".lua\"")
			if not s then
				SYSError("Deleting rule "..v.." error.")
			else
				SYSDebug("Rule "..v.." successfully deleted.")
			end
		else
			SYSError("Non string value in delete list with index #"..i)
		end
	end
	os.execute("rm -Rf "..config.spool_dir..uuid)
end

----------------------------------------------
-- Executes rules
-- @param rule string
-- @param output bool
-- @return table
----------------------------------------------
function execute_rule(rule, output)
	local passive = true
	local output_file = io.stdout
	local result = {};
	local result_top = {};
	result.items = {}
	result_top.items = {}

	if (rule == "all") then
		status, res, msg = pcall(get_rule_list)
	else
		res = { rule }
		status = true
	end
	if output then
		passive = false
	end
	if status then
		for i, rule in ipairs(res) do
			local time_mem = socket.gettime()
			local loop_start = time_mem

			-- this rule loadfile must MUST set the variable "script"
			local _, res, msg = pcall(loadfile, config.tree_dir..rule..".lua")

			SYSDebug(string.format("Load rule complete. %.4f sec", socket.gettime() - time_mem))
			time_mem = socket.gettime()
			if res == nil then
				SYSError("Could not load rule file for rule "..rule..": ".. msg)
				error()
			else
				res()
				SYSDebug(string.format("Execution of rule complete. %.4f sec", socket.gettime() - time_mem))
			end
			time_mem = socket.gettime()

			local exec_s, r, m = execute(script, "", passive)
			if exec_s then
				SYSDebug(string.format("Rule processing complete. %.4f sec", socket.gettime() - time_mem))
			else
				SYSError("Execution of rule failed: "..m)
			end

			-- Calculate full tree results
			status, res, msg = pcall(prepare_full, script)
			if not status then
				SYSError("Json prepare failed: "..(res or msg) or "")
				error()
			elseif res then
				result.items[rule] = res
			end

			-- If materialized, send command to nagios
			if res["materialized"] and not output then
				mhost = host(res.materialized.host)
				if mhost then
					mserv = mhost:service(res.name)
					if mserv then
						SYSDebug("["..os.time().."] Last check = "..mserv.last_check.." Last state change = "..mserv.last_state_change)
						-- Check if last check changed the state
						if mserv.last_check == mserv.last_state_change and config.save_report_data then
							local parsed_data = {
								data = ""
							}
							-- insert report data
							if res.result.status ~= 0 then
								parsed_data.data = parse_report_data(res)
							end
							parsed_data["meta"] = {
								timestamp = mserv.last_state_change,
								host = res.materialized.host,
								name = res.name,
								state = res.result.status
							}
							-- TODO: Implement popen2 to be able to read stderr on failure.
							if parsed_data.data ~= false then
								local handle = assert(io.popen("/usr/bin/php /opt/monitor/op5/ninja/index.php synergy_report/insert_report_data", "w"))
								if not handle then
									SYSError("Could not open file handle to synergy_report")
								else
									handle:write(json.encode(parsed_data))
									handle:close()
								end
							else
								SYSError("Trying to process report data for an empty rule. Skipping "..(res.name or ""))
							end
						end
					end
				else
					SYSError("Could not find host for materialized service. Ignoring materialization of "..(res.materialized.host or "N/A"))
				end
			end

			-- Generate top level results for widget
			local status, topresult, msg = pcall(prepare_top, script)
			if status then
				result_top.items[rule] = topresult
				SYSDebug("Top result for rule "..rule.." generated")
			else
				SYSDebug("Prepare top json for rule "..rule.." failed")
			end

			SYSDebug(string.format("Rule executed in %.4f sec", socket.gettime() - loop_start))
		end
	else
		SYSError("Failed getting rule list: "..(res or msg) or "")
	end

	cached_full = json.encode(result)
	cached_top = json.encode(result_top)

	return r
end

----------------------------------------------
-- Format calculated result to full tree
-- @param obj table
-- @return table
----------------------------------------------
function prepare_full(obj)
	l = l or 0
	local result = {}
	local c = 0
	result.name = obj["name"]
	result.type = obj["intent"]["type"]
	result.rule = obj["intent"]["name"]
	result.args = {}
	for k,v in ipairs(obj["intent"]["args"] or {}) do
		table.insert(result.args, v)
	end
	result["materialized"] = obj["materialized"] or nil
	if type(obj["components"]) == "table" then
		result.items = {}
		for k,v in ipairs(obj["components"]) do
			local itemname = v.name
			result.items[itemname] = prepare_full(v)
		end
	end
	if type(obj["result"]) ~= "table" or table.getn(obj["result"]) == 0 then
		result.status = "UNKNOWN"
		result.result = {}
		result.result.msg = "No result"
		result.result.status = 3
	else
		result.status = status_map[obj["result"][1]["status"] + 1]
		result.result = {}
		for k,v in pairs(obj["result"][1]) do
			result.result[k] = v
		end
	end
	return result
end

----------------------------------------------
-- Format calculated top level result
-- @param obj table
-- @return table
----------------------------------------------
function prepare_top(obj)
	SYSDebug("Preparing top json "..obj["name"])
	local res = {}
	res["name"] = obj["name"]
	res["type"] = obj["intent"]["type"]
	res["rule"] = obj["intent"]["name"]
	res["args"] = obj["intent"]["args"]
	res["materialized"] = obj["materialized"]
	if type(obj["result"]) ~= "table" or table.getn(obj["result"]) == 0 then
		res["result"] = {status = "UNKNOWN", result = {msg = "no result", status = 3}}
	else
		res["result"] = obj["result"][1]
	end
	return res
end

----------------------------------------------
-- Strips full result to necessary data for reports
-- @param data table
-- @return table
-- @TODO Obviously this is not the way to do this.
-- Fomrmat the data from the object instead.
----------------------------------------------
function parse_report_data(data)
	if not data.items then
		-- Tree might be empty, return false.
		return false
	end
	for k,v in pairs(data.items) do
		if v.items then
			v.status = v.result.status
			v.result.status = nil
			v.type = nil
			v.args = nil
			v.result.last_check = nil
			v.rule = nil
			data.items[k] = parse_report_data(v)
		elseif v.status == "OK" then
			-- Remove Ok objects
			data.items[k] = nil
		else
			-- Strip data that's not needed
			v.rule = nil
			v.type = nil
			v.status = v.result.status
			v.result = nil
			v.args = nil
		end
	end
	return data
end

----------------------------------------------
-- Write passive check to nagios command pipe
-- @param cmd string
-- @return void
----------------------------------------------
function passive_check_proccess(cmd)
	local pipe, msg = io.open(config.nagios_cmd_path, "a+")
	pipe:write(cmd)
	pipe:close()
end

----------------------------------------------
-- Format passive check result
-- @param host_name string
-- @param service_dsc string
-- @param code int
-- @param msg string
-- @return void
----------------------------------------------
function passive_check(host_name, service_dsc, code, msg)
	local cmd = string.format("[%d] PROCESS_SERVICE_CHECK_RESULT;%s;%s;%d;%s;\n", os.time(), host_name, service_dsc, code, msg)
	local exit = false
	local status, res, msg
	local tries = 0
	while not exit  and tries < config.max_retries do
		status, res, msg = pcall(passive_check_proccess, cmd)
		if status then
			exit = true
		else
			tries = tries + 1
			SYSError("Passive check: "..cmd.." failed. Error: "..(msg or ""))
			SYSDebug("Trying again. "..tries.."/"..config.max_retries)
			sleep(config.wait_on_fail)
		end
	end
end

----------------------------------------------
-- Fetch all trees as tables
-- @return table
----------------------------------------------
function get_rule_list()
	local res = {}
	for file in lfs.dir(config.tree_dir) do
		if (file ~= "." and file ~= ".." and file:sub(-4) == ".lua") then
			table.insert(res, file:sub(0, -5))
		end
	end
	return res
end

----------------------------------------------
-- Intersect contacts from contactgroups
-- @param list table
-- @return string
----------------------------------------------
function intersection(list, as_json)
	if #list == 0 then
		return {}
	end
	local list_count = 0
	local items = {}
	for k,v in ipairs(list) do
		if #v == 0 then
			return {}
		end
		if (v ~= "*") and (not ((#v == 1) and (v[1] == "*"))) then
			list_count = list_count + 1
			local used = {}
			for sk, sv in ipairs(v) do
				if not used[sv] then
					table.insert(items, sv)
					used[sv] = true
				end
			end
		end
	end
	if #items == 0 then
		return "*"
	end
	table.sort(items)
	if as_json then
		print(json.encode(items))
	end
	local last_v
	local last_l = 0
	local res = {}
	for k,v in ipairs(items) do
		if last_v == v then
			last_l = last_l + 1
		else
			if last_l == list_count then
				table.insert(res, last_v)
			end
			last_l = 1
			last_v = v
		end
	end
	if last_l == list_count then
		table.insert(res, last_v)
	end
	if as_json then
		print(json.encode(res))
	end
	return res
end

----------------------------------------------
-- Sets states, contacts on objects
-- @param obj table
-- @param path string
-- @param passive bool
-- @return table
----------------------------------------------
function execute(obj, path, passive)
	local new_path = path.."/"..obj.name
	local arguments = {}
	local service_contacts
	local host_contacts
	if type(obj.components) == "table" then
		host_contacts = {}
		service_contacts = {}
		for i,v in ipairs(obj.components) do
			local res = execute(v, new_path, passive)
			if type(res) == "table" then
				arguments = table_merge(arguments, res)
				local serv_cont = res[1].service_contacts
				serv_cont = table_merge(serv_cont, contactgroups_get_contacts(res[1].service_contact_groups))
				table.insert(service_contacts, serv_cont)
				local host_serv = res[1].host_contacts
				host_serv = table_merge(host_serv, contactgroups_get_contacts(res[1].host_contact_groups))
				table.insert(host_contacts, host_serv)
			end
		end
		local s, m
		s, service_contacts, m = pcall(intersection, service_contacts, as_json)
		if not s then
			print("Couldn't intersect service contacts: "..m)
		end
		s, host_contacts, m = pcall(intersection, host_contacts)
		if not s then
			print("Couldn't intersect host contacts: "..m)
		end
	end

	if type(obj.intent) == "table" then
		local status, msg, lsc
		obj.result = obj.intent.func(arguments, unpack(obj.intent.args))
		if obj.materialized then
			-- Object is materialized, fetch last state change from the materialized service.
			mhost = host(obj.materialized.host)
			if mhost then
				mserv = mhost:service(obj.name)
				if mserv then
					lsc = mserv.last_state_change
				else
					-- Materialized service not found, this is bad! Automagikz has entered the building.
					local s, r, m = pcall(materialize_service, obj.materialized.host, obj.name)
					if not s then
						-- Materialization failed. Die horribly!
						SYSError("Failed to materialize service:"..obj.materialized.host..";"..obj.name)
--						os.exit(1)
					end
				end
				if passive then
					passive_check(obj.materialized.host, obj.name, obj.result[1].status, obj.result[1].msg)
				end
			else
				SYSError("Could not find host for materialized service. Ignoring materialization of "..(obj.materialized.host or "N/A"))
			end
		end
		if(obj.intent.type ~= "nagios") then
			local res = obj.result[1]
			if type(states[new_path]) == "table" then
				if states[new_path].status ~= res.status then
					states[new_path].status = res.status
					-- Set last state change
					if lsc then
						states[new_path].lsc = lsc
					else
						res.last_state_change = os.time()
					end
				else
					res.last_state_change = states[new_path].lsc
				end
			else
				states[new_path] = {}
				states[new_path].status = res.status
				-- Set last state change
				if lsc then
					states[new_path].lsc = lsc
				else
					states[new_path].lsc = os.time()
				end
				res.last_state_change = states[new_path].lsc
			end
			res.last_check = os.time()
		end
		if type(obj.components) == "table" then
			obj.result[1].service_contacts = service_contacts
			obj.result[1].host_contacts = host_contacts
		end
		if not obj.result[1].host_contacts then
			obj.result[1].host_contacts = "*"
		end
		if not obj.result[1].service_contacts then
			obj.result[1].service_contacts = "*"
		end
		return obj.result
	end
end

----------------------------------------------
-- Materializes not materialized service using nacoma API
-- @param host string
-- @param service string
-- @return void
----------------------------------------------
function materialize_service(host, service)
	local url = "/monitor/index.php/synergy/view#/tree:"
	SYSDebug("Materializing service: "..host..";"..service)
	local materialized = os.execute("php "..config.nacoma_apiurl.." -t service -o template=bizproc_service -o host_name=\""..host.."\" -o service_description=\""..service.."\" -o action_url=\""..url..service.."\" -u synergy_daemon")
	if materialized == 0 then
		local save = os.execute("php "..config.nacoma_apiurl.." -a save_config -u synergy_daemon")
		if save == 0 then
			return true
		else
			error("Failed to save config when creating service: "..host..";"..service)
			return false
		end
	else
		error("Failed to create service: "..host..";"..service)
	end
	return false
end

----------------------------------------------
-- Packs all arguments into table
-- @param ...
-- @return table
----------------------------------------------
function pack_arguments(...)
	local r = {}
	for k,v in ipairs(arg) do
		if k ~= "n" and v then
			table.insert(r, v)
		end
	end
	return r
end

----------------------------------------------
-- Get contacts from contactgroups
-- @param groups table
-- @return table
----------------------------------------------
function contactgroups_get_contacts(groups)
	local res = {}
	if not groups then
		groups = "*"
	end
	if groups == "*" then
	return "*"
	end
	for i,name in ipairs(groups) do
		if not contactgroups_cache[name] then
			contactgroups_cache[name] = response_to_obj(mklive:request(Request.new("contactgroups"):where("name = ?", name))).members
		end
		res = table_merge(res, contactgroups_cache[name])
	end
	return res
end

----------------------------------------------
-- Merge tables
-- @param dst table
-- @param src table
-- @return table
----------------------------------------------
table_merge = function(dst, src)
	if src == "*" then
		if dst == "*" then
			return "*"
		end
		table.insert(dst, "*")
		return dst
	end
	for i,v in ipairs(src) do
		table.insert(dst, v)
	end
	return dst
end
